"""
OMVS Module Clients for ODPi Egeria.
"""
from pyegeria.omvs.action_author import ActionAuthor
from pyegeria.omvs.actor_manager import ActorManager
from pyegeria.omvs.asset_catalog import AssetCatalog
from pyegeria.omvs.asset_maker import AssetMaker
from pyegeria.omvs.automated_curation import AutomatedCuration
from pyegeria.omvs.classification_explorer import ClassificationExplorer
from pyegeria.omvs.collection_manager import CollectionManager
from pyegeria.omvs.community_matters_omvs import CommunityMatters
from pyegeria.omvs.data_designer import DataDesigner
from pyegeria.omvs.data_discovery import DataDiscovery
from pyegeria.omvs.data_engineer import DataEngineer
from pyegeria.omvs.digital_business import DigitalBusiness
from pyegeria.omvs.external_links import ExternalReferences
from pyegeria.omvs.feedback_manager import FeedbackManager
from pyegeria.omvs.full_omag_server_config import FullServerConfig
from pyegeria.omvs.glossary_manager import GlossaryManager
from pyegeria.omvs.governance_officer import GovernanceOfficer
from pyegeria.omvs.lineage_linker import LineageLinker
from pyegeria.omvs.location_arena import LocationArena
from pyegeria.omvs.metadata_expert import MetadataExpert
from pyegeria.omvs.metadata_explorer_omvs import MetadataExplorer
from pyegeria.omvs.my_profile import MyProfile
from pyegeria.omvs.notification_manager import NotificationManager
from pyegeria.omvs.people_organizer import PeopleOrganizer
from pyegeria.omvs.product_manager import ProductManager
from pyegeria.omvs.project_manager import ProjectManager
from pyegeria.omvs.reference_data import ReferenceDataManager
from pyegeria.omvs.registered_info import RegisteredInfo
from pyegeria.omvs.runtime_manager import RuntimeManager
from pyegeria.omvs.schema_maker import SchemaMaker
from pyegeria.omvs.server_operations import ServerOps
from pyegeria.omvs.solution_architect import SolutionArchitect
from pyegeria.omvs.specification_properties import SpecificationProperties
from pyegeria.omvs.subject_area import SubjectArea
from pyegeria.omvs.template_manager_omvs import TemplateManager
from pyegeria.omvs.time_keeper import TimeKeeper
from pyegeria.omvs.valid_metadata import ValidMetadataManager
from pyegeria.omvs.valid_metadata_lists import ValidMetadataLists
from pyegeria.omvs.valid_type_lists import ValidTypeLists

__all__ = [
    "ActionAuthor",
    "ActorManager",
    "AssetCatalog",
    "AssetMaker",
    "AutomatedCuration",
    "ClassificationExplorer",
    "CollectionManager",
    "CommunityMatters",
    "DataDesigner",
    "DataDiscovery",
    "DataEngineer",
    "DigitalBusiness",
    "ExternalReferences",
    "FeedbackManager",
    "FullServerConfig",
    "GlossaryManager",
    "GovernanceOfficer",
    "LineageLinker",
    "LocationArena",
    "MetadataExpert",
    "MetadataExplorer",
    "MyProfile",
    "NotificationManager",
    "PeopleOrganizer",
    "ProductManager",
    "ProjectManager",
    "ReferenceDataManager",
    "RegisteredInfo",
    "RuntimeManager",
    "SchemaMaker",
    "ServerOps",
    "SolutionArchitect",
    "SpecificationProperties",
    "SubjectArea",
    "TemplateManager",
    "TimeKeeper",
    "ValidMetadataManager",
    "ValidMetadataLists",
    "ValidTypeLists",
]
